/* kanji.c */

/* Author:
 *	Junichiro ITOH
 *	itojun@foretune.co.jp / itojun@mt.cs.keio.ac.jp / JINnet itojun
 */

/* This file contains some functions to handle kanji characters */
/* internal code: EUC */


/* prototype definition for lseek() needed */
#if MSDOS
# include <io.h>
#else
extern	long	lseek();
#endif
#include "config.h"
#include "ctype.h"
#include "vi.h"
#if KANJI
#include "kanji.h"
#ifndef SEEK_SET
# define SEEK_SET	0
# define SEEK_CUR	1
# define SEEK_END	2
#endif


/*----------------------------------------------------------------------*/

unsigned int	getkanji(text)
	unsigned char	*text;
{
	return ((unsigned int)text[0] << 8) | text[1];
}


unsigned int	localtojis(ch)
	unsigned int	ch;
{
	return ch & 0x7f7f;
}


unsigned int	jistolocal(ch)
	unsigned int	ch;
{
	return ch | 0x8080;
}


unsigned int	localtosjis(ch)
	unsigned int	ch;
{
	int	high;
	int	low;
	
	high = (ch >> 8) & 0x7f;
	low = ch & 0x7f;
	if (high & 1)
		low += 0x1f;
	else
		low += 0x7d;
	if (low >= 0x7f)
		low++;
	high = (high - 0x21 >> 1) + 0x81;
	if (high > 0x9f)
		high += 0x40;
	return ((unsigned int)high << 8) | low;
}


unsigned int	sjistolocal(ch)
	unsigned int	ch;
{
	int	high;
	int	low;
	
	high = (ch >> 8) & 0xff;
	low = ch & 0xff;
	if (high <= 0x9f)
		high -= 0x71;
	else
		high -= 0xb1;
	high = high * 2 + 1;
	if (low > 0x7f)
		low--;
	if (low >= 0x9e)
	{
		low -= 0x7d;
		high++;
	}
	else
	{
		low -= 0x1f;
	}
	return (((unsigned int)high << 8) | low) | 0x8080;
}


unsigned int	localtoeuc(ch)
	unsigned int	ch;
{
	return ch;
}


unsigned int	euctolocal(ch)
	unsigned int	ch;
{
	return ch;
}

/*----------------------------------------------------------------------*/

int	kanji_forward(str, pos, n)
	char	*str;	/* NULL-terminated string */
	int	pos;	/* current cursor position */
	int	n;	/* forward n characters */
{
	REG int		i;
	REG int		j;

	j = 0;
	for (i = 0; i < n; i++)
	{
		if (!str[pos + j])
			return j + 1;		/* force error */
		if (iskanji(str[pos + j]))
		{
			if (!str[pos + j + 1])
				return j + 2;	/* force error */
			else
				j += 2;
		}
		else
		{
			j++;
		}
	}
	return j;
}


int	kanji_backward(str, pos, n)
	char	*str;	/* NULL-terminated string */
	int	pos;	/* current cursor position */
	int	n;	/* backward n characters */
{
	REG int		i;
	REG int		j;

	j = 0;
	for (i = 0; i < n; i++)
	{
		if (pos < j)
			return pos + 1;		/* force error */
		j++;
		if (kanji_fix(str, pos - j))
			j++;
	}
	return j;
}


/* Special thanks to serow */
int	kanji_fix(str, pos)
	char	*str;	/* NULL-terminated string */
	int	pos;	/* current cursor position */
{
	REG int		i;
	REG int		count;

	if (pos <= 0)
		return 0;
	i = pos - 1;
	count = 0;
	while (iskanji(str[i]) && 0 <= i)
	{
		i--;
		count++;
	}
	if (count % 2) 
		return 1;
	else
		return 0;
}

/*----------------------------------------------------------------------*/

static
int	kanji_pack(first, second)
	int	first;
	int	second;
{
	return ((unsigned int)first << 8) | (second & 0xff);
}


int	kanji_ispack(key)
	int	key;
{
	return key & 0xff00;
}


int	kanji_first(key)
	unsigned int	key;
{
	return (unsigned int)key >> 8;
}


int	kanji_second(key)
	unsigned int	key;
{
	return key & 0xff;
}


int	kanji_getkey(when)
	int	when;
{
	int	key;

	key = getkey(when);
	if (iskanji(key))
		key = kanji_pack(key, getkey(when));
	return key;
}


int	kanji_match(str, pos, key)
	char	*str;	/* NULL-terminated string */
	int	pos;	/* current cursor position */
	int	key;	/* character to search */
{
	if (kanji_ispack(key))
	{
		if ((kanji_first(key) & 0xff) == (str[pos] & 0xff)
		 && (kanji_second(key) & 0xff) == (str[pos+1] & 0xff))
			return TRUE;
		else
			return FALSE;
	}
	else
	{
		return str[pos] == key;
	}
}

/*----------------------------------------------------------------------*/

	/* kanji character type start code */
#define	K_SPACE		0x2121
#define	K_ALPHABETTE	0x2330
#define	K_HIRAGANA	0x2421
#define	K_KATAKANA	0x2521
#define	K_GREEK		0x2621
#define	K_CHINESE	0x3021

	/* ranks of character type */
#define	R_MARKS		0
#define	R_ALPHABETTE	5
#define	R_HIRAGANA	1
#define	R_KATAKANA	10
#define	R_GREEK		5
#define	R_CHINESE	20
#define	R_NONKANJI	100

#define	arraysize(x)		(sizeof(x)/sizeof(*(x)))

	/* special characters */
static
unsigned int	sp_china[] = {		/* VWXY */
	0x2137, 0x2138, 0x2139, 0x213a,
};
static
unsigned int	sp_kana[] = {		/* RSTU[` */
	0x2133, 0x2134, 0x2135, 0x2136, 0x213c, 0x2141,
};


static
int	isthere(ch, class, size)
	unsigned int	ch;
	unsigned int	*class;
	int		size;
{
	REG int		i;

	for (i = 0; i < size; i++)
	{
		if (ch == class[i])
			return 1;
	}
	return 0;
}


int	kanji_chartype(text, type)
	unsigned char	*text;
	int		type;	/* default type of character */
{
	int		ret;
	unsigned int	ch;

	if (iskanji(text[0]) && text[1])	/* kanji */
	{
		ch = localtojis(getkanji(text));
		if (isthere(ch, sp_china, arraysize(sp_china)))
				return R_CHINESE;
		if (isthere(ch, sp_kana, arraysize(sp_kana)))
		{
			if (type == 1)	/* hiragana */
				return R_HIRAGANA;	/* hiragana */
			else		/* not hiragana */
				return R_KATAKANA;	/* katakana */
		}
		if (ch < K_ALPHABETTE)		/* marks */
			ret = R_MARKS;
		else if (ch < K_HIRAGANA)	/* alphabette */
			ret = R_ALPHABETTE;
		else if (ch < K_KATAKANA)	/* hiragana */
			ret = R_HIRAGANA;
		else if (ch < K_GREEK)		/* katakana */
			ret = R_KATAKANA;
		else if (ch < K_CHINESE)	/* greek, russian */
			ret = R_GREEK;
		else				/* chinese */
			ret = R_CHINESE;
	}
	else	/* not kanji */
	{
		ret = R_NONKANJI;
	}
	return ret;
}


int	kanji_isspace(text)
	unsigned char	*text;
{
	unsigned int	ch;
	int		ret;

	ret = FALSE;
	if (iskanji(text[0]) && text[1])
	{
		ch = localtojis(getkanji(text));

		if (ch == K_SPACE)
			ret = TRUE;
	}

	return ret;
}


#ifndef	NO_SENTENCE
	/* kanji periods */
static
unsigned int	sp_period[] = {		/* BDHI */
	0x2123, 0x2125, 0x2129, 0x212a,
};


int	kanji_isperiod(text)
	unsigned char	*text;
{
	unsigned int	ch;
	int		ret;

	ret = FALSE;
	if (iskanji(text[0]) && text[1])
	{
		ch = localtojis(getkanji(text));

		if (isthere(ch, sp_period, arraysize(sp_period)))
		{
			ret = TRUE;
			if (iskanji(text[2]) && text[3])
			{
				ch = localtojis(getkanji(text+2));
				if (isthere(ch, sp_period, arraysize(sp_period)))
					ret = FALSE;
			}
		}
	}
	
	return ret;
}
#endif	/*NO_SENTENCE*/

/*----------------------------------------------------------------------*/

/* recognized escape sequences for input:
 *	kanji in	ESC-$-@	ESC-$-B
 *	ascii in	ESC-(-J ESC-(-B
 */

/*specify escape sequence for output*/
static	char	*kanjiIn	= "\033$@";
static	char	*kanjiOut	= "\033(J";


int	buf2ext(dest, src, slen, code)
	char	*dest;
	char	*src;
	int	slen;
	char	*code;
{
	int	d;
	int	s;
	int	i;
	int	kflag;
	int	pack;

	s = d = 0;
	kflag = FALSE;
	while (s < slen)
	{
		if (iskanji(src[s]) && s+1 < slen)
		{
			if (code[0] == 's')	/*sjis*/
			{
				pack = localtosjis(getkanji(&src[s]));
				dest[d++] = kanji_first(pack);
				dest[d++] = kanji_second(pack);
				s += 2;
			}
			else if (code[0] == 'j')	/*jis*/
			{
				if (!kflag)
				{
					i = 0;
					while (kanjiIn[i])
						dest[d++] = kanjiIn[i++];
				}
				pack = localtojis(getkanji(&src[s]));
				dest[d++] = kanji_first(pack);
				dest[d++] = kanji_second(pack);
				s += 2;
			}
			else	/*euc*/
			{
				pack = localtoeuc(getkanji(&src[s]));
				dest[d++] = kanji_first(pack);
				dest[d++] = kanji_second(pack);
				s += 2;
			}
			kflag = TRUE;
		}
		else
		{
			if (kflag && code[0] == 'j')	/*jis*/
			{
				i = 0;
				while (kanjiOut[i])
					dest[d++] = kanjiOut[i++];
			}
			kflag = FALSE;
			dest[d++] = src[s++];
		}
	}
	if (kflag && code[0] == 'j')	/*jis*/
	{
		i = 0;
		while (kanjiOut[i])
			dest[d++] = kanjiOut[i++];
	}
	dest[d] = '\0';
	return d;
}


int	ext2buf(dest, src, slen, code)
	char	*dest;
	char	*src;
	int	slen;
	char	*code;
{
	int	d;
	int	s;
	int	i;
	int	kflag;
	int	pack;

	s = d = 0;
	kflag = FALSE;
	while (s < slen)
	{
#if !KANJI_OPTIMIST
		if (code[0] == 'j' && src[s] == '\033')
#else
		if (src[s] == '\033')
#endif
		{
			if (src[s+1] == '$' && (src[s+2] == '@' || src[s+2] == 'B') && s+2 < slen)
			{
				s += 3;
				kflag = TRUE;
				continue;
			}
			else if (src[s+1] == '(' && (src[s+2] == 'B' || src[s+2] == 'J') && s+2 < slen)
			{
				s += 3;
				kflag = FALSE;
				continue;
			}
		}

#if !KANJI_OPTIMIST
		if (code[0] == 'j' && kflag && isjiskanji(src[s]) && s+1 < slen)
#else
		if (kflag && isjiskanji(src[s]) && s+1 < slen)
#endif
		{
			pack = jistolocal(getkanji(&src[s]));
			dest[d++] = kanji_first(pack);
			dest[d++] = kanji_second(pack);
			s += 2;
		}
		else if (code[0] == 's' && issjiskanji(src[s]) && s+1 < slen)
		{
			pack = sjistolocal(getkanji(&src[s]));
			dest[d++] = kanji_first(pack);
			dest[d++] = kanji_second(pack);
			s += 2;
		}
		else if (iseuckanji(src[s]) && s+1 < slen)
		{
			pack = euctolocal(getkanji(&src[s]));
			dest[d++] = kanji_first(pack);
			dest[d++] = kanji_second(pack);
			s += 2;
		}
		else
		{
			dest[d++] = src[s++];
		}
	}
	dest[d] = '\0';
	return d;
}

/*----------------------------------------------------------------------*/

#if KANJI_DETECT
char	*kanji_detect(fd)
	int	fd;
{
	int	len;
	long	pos;
	char	*code;
	int	offset;
	int	i;

	pos = lseek(fd, 0L, SEEK_CUR);

	offset = 0;
	code = (char *)0;
	while (code == (char *)0)
	{
		len = read(fd, tmpblk.c + offset, BLKSIZE - offset);
		if (len <= 0)
			break;
		len += offset;
		offset = 0;
		i = 0;

		while (i < len)
		{
			if (len - i < 3)
			{
				memcpy(tmpblk.c + i, tmpblk.c, len - i);
				offset = len - i;
				break;
			}

			if (tmpblk.c[i] == '\033')
			{
				if ((tmpblk.c[i+1] == '$' && (tmpblk.c[i+2] == '@' || tmpblk.c[i+2] == 'B'))
				 || (tmpblk.c[i+1] == '(' && (tmpblk.c[i+2] == 'J' || tmpblk.c[i+2] == 'B')))
				{
					code = "jis+";
					goto breakBreak;
				}
				else
				{
					i++;
				}
			}
			else
			{
				switch (((iseuckanji(tmpblk.c[i]))?1:0)
				       +((issjiskanji(tmpblk.c[i]))?2:0))
				{
				case 0:
					i++;
					break;

				case 1:
					code = "euc+";
					goto breakBreak;
					/*NOTREACHED*/

				case 2:
					code = "sjis+";
					goto breakBreak;
					/*NOTREACHED*/

				case 3:
					i += 2;
					break;
				}
			}
		}
	}
breakBreak:

	lseek(fd, pos, SEEK_SET);
	return code;
}
#endif	/*KANJI_DETECT*/

/*----------------------------------------------------------------------*/

int	ktwrite(fd, buf, len, code)
	int	fd;
	char	*buf;
	int	len;
	char	*code;
{
	char	*tmp;
	int	tlen;
	int	ret;

	tmp = (char *)malloc(len * 2 * sizeof(char));
	tlen = buf2ext(tmp, buf, len, code);
	ret = twrite(fd, tmp, tlen);
	free(tmp);
	return ret;
}


int	ktread(fd, buf, len, code)
	int	fd;
	char	*buf;
	int	len;
	char	*code;
{
	char	*tmp;
	int	tlen;
	int	i;
	int	ret;
	long	pos;

	tmp = (char *)malloc(len * 2 * sizeof(char));
	pos = lseek(fd, 0L, SEEK_CUR);
	tlen = tread(fd, tmp, len);
	if (tlen <= 0)
	{
		free(tmp);
		return tlen;
	}

	i = tlen;
	while (0 <= --i)
	{
		if (tmp[i] == '\n')
			break;
	}

	if (i < 0)
	{
		ret = ext2buf(buf, tmp, tlen, code);
	}
	else
	{
		lseek(fd, pos, SEEK_SET);
		tlen = tread(fd, tmp, i + 1);
		ret = ext2buf(buf, tmp, tlen, code);
	}
	free(tmp);
	return ret;
}


int	ktread_slow(fd, buf, len, code)
	int	fd;
	char	*buf;
	int	len;
	char	*code;
{
	char	*tmp;
	int	tlen;
	int	i;
	int	ret;

	tmp = (char *)malloc(len * 2 * sizeof(char));
	i = 0;
	while (i < len)
	{
		tlen = tread(fd, &tmp[i++], 1);
		if (tlen <= 0)
		{
			if (i == 1)	/*first char*/
			{
				free(tmp);
				return tlen;
			}
			else
			{
				break;
			}
		}

		if (tmp[i-1] == '\n')
			break;
	}

	ret = ext2buf(buf, tmp, i, code);
	free(tmp);
	return ret;
}


int	kttywrite(buf, len)
	char	*buf;
	int	len;
{
	char	*tmp;
	int	tlen;
	int	ret;

	tmp = (char *)malloc(len * 2 * sizeof(char));
	tlen = buf2ext(tmp, buf, len, o_kdisplaycode);
	ret = ttywrite(tmp, tlen);
	free(tmp);
	return ret;
}


int	kttyread(buf, len, time)
	char	*buf;
	int	len;
	int	time;
{
static	char	preserve[KEYBUFSIZE]	= "";
static	int	kflag		= FALSE;
static	int	pack		= 0;
static	int	sequence	= 0;
	int	ret;
	int	i;
	int	j;

	do
	{
		if (sequence || pack)
			ret = ttyread(buf, len, o_keytime[0]);
		else
			ret = ttyread(buf, len, time);
		i = 0;
		j = strlen(preserve);
		
		/*timeout*/
		if (!ret)
		{
			if (sequence)
			{
				preserve[j++] = '\033';
				if (sequence == 2)
					preserve[j++] = '$';
				else if (sequence == 3)
					preserve[j++] = '(';
				sequence = 0;
			}
			else if (pack)
			{
				preserve[j++] = pack;
				pack = 0;
			}
			preserve[j] = '\0';
			break;
		}
		
		/*parse incoming keystroke*/
		while (i < ret)
		{
			if (sequence)
			{
				if (sequence == 1)	/*ESC found*/
				{
					if (buf[i] == '$')
					{
						i++;
						sequence = 2;
					}
					else if (buf[i] == '(')
					{
						i++;
						sequence = 3;
					}
					else
					{
						preserve[j++] = '\033';
						preserve[j++] = buf[i++];
						sequence = 0;
					}
				}
				else if (sequence == 2)		/*ESC-$ found*/
				{
					if (buf[i] == '@' || buf[i] == 'B')
					{
						kflag = TRUE;
						i++;
					}
					else
					{
						preserve[j++] = '\033';
						preserve[j++] = '$';
						preserve[j++] = buf[i++];
					}
					sequence = 0;
				}
				else if (sequence == 3)		/*ESC-( found*/
				{
					if (buf[i] == 'B' || buf[i] == 'J')
					{
						kflag = FALSE;
						i++;
					}
					else
					{
						preserve[j++] = '\033';
						preserve[j++] = '(';
						preserve[j++] = buf[i++];
					}
					sequence = 0;
				}
				else				/*ambigous state*/
				{
					sequence = 0;
				}
			}
			else if (pack)
			{
				pack = kanji_pack(pack, buf[i++]);
				if (o_kinputcode[0] == 'j')
					pack = jistolocal(pack);
				else if (o_kinputcode[0] == 's')
					pack = sjistolocal(pack);
				else if (o_kinputcode[0] == 'e')
					pack = euctolocal(pack);
				preserve[j++] = kanji_first(pack);
				preserve[j++] = kanji_second(pack);
				pack = 0;
			}
			else
			{
				if (o_kinputcode[0] == 'j' && buf[i] == '\033')
				{
					i++;
					sequence = 1;
				}
				else if ((o_kinputcode[0] == 'j' && kflag && isjiskanji(buf[i]))
					|| (o_kinputcode[0] == 's' && issjiskanji(buf[i]))
					|| (o_kinputcode[0] == 'e' && iseuckanji(buf[i])))
				{
					pack = buf[i++];
				}
				else
				{
					preserve[j++] = buf[i++];
				}
			}
			preserve[j] = '\0';
		}
	} while (sequence || pack || (!time && !j));
	
	strcpy(buf, preserve);
	preserve[0] = '\0';
	return strlen(buf);
}

#endif	/*KANJI*/

