/* $Id: xargs.c 1.5 91/03/10 09:08:26 serow Exp Locker: serow $
 * wild card expand routine for Turbo-C, MS-C, LSI-C..
 * This code is in the public domain.
 *
 * $Log:	xargs.c $
 * Revision 1.5  91/03/10  09:08:26  serow
 * glob_filename() interface
 * CMDLINE environ
 * 
 * Revision 1.4  91/02/23  19:41:34  serow
 * error, missing }, ]
 * 
 * Revision 1.3  91/02/22  06:58:50  serow
 * globbing by AssistantIO
 * 
 */

#include <dos.h>
#include <string.h>
#include <io.h>
#include <stdlib.h>

#ifndef GLOB
#define GLOB	1	/* NON 0: enable globbing */
#endif
#ifndef DOT_DLM
#define DOT_DLM	0	/* NON 0: treat '.' as str separator when globbing */
#endif
#ifndef HIDDEN_DOT
#define HIDDEN_DOT	1	/* NON 0: treat '.' & '..' as hidden file */
#endif
#ifndef ArgsAttr
#define ArgsAttr	1	/* Default attribute when file search */
#endif
#ifndef DO_SORT
#define DO_SORT	1	/* NON 0:wild card args will be sorted by name */
#endif
#ifndef CVLOWER
#define CVLOWER	1	/* NON 0: convert filenames to lower case */
#endif

#define PSP	_psp

#if GLOB
#define L_CURLY	'{'
#define R_CURLY	'}'
#if DOT_DLM
#define is_end(c)	((c) == '\0' || (c) == '.')
#else
#define is_end(c)	((c) == '\0')
#endif

#define IGNORE_CASE	1

#define iswildchar(c) ((c)=='*' || (c)=='?' || (c)=='[' || (c)==L_CURLY)
#define Isupper(c)	('A' <= (c) && (c) <= 'Z')
#define Tolower(c)	((c) | 0x20)
#else
#define iswildchar(c) ((c)=='*' || (c)=='?')
#endif /* !GLOB */

#define Isspace(c)	(c == ' ' || c == '\t' || c == '\n' )
#define UCH(c)  ((unsigned char)(c))
#define isk1(c) ((0x81<=UCH(c)&&UCH(c)<=0x9F)||(0xE0<=UCH(c)&&UCH(c)<=0xFC))
#define isk2(c) ((0x40<=UCH(c)&&UCH(c)<=0x7E)||(0x80<=UCH(c)&&UCH(c)<=0xFC))

#define REALLOC_UNIT	128

#ifdef LDATA
#undef LDATA
#endif

#ifdef __TURBOC__
#if (sizeof(char *) == 4)
#define LDATA	1
#endif
#else /* may be MS-C */
#if defined(M_I86LM) || defined(M_I86CM) || defined(M_I86HM)
#define LDATA	1
#endif
#endif

#if defined(LSI_C)
/* union REGS dose not have '.cflag' but '.flags' */
#else
#define HAVE_CFLAG	1
#endif

#ifdef _MSC_VER
/* MS-C */
#define NEAR	_near
#else
#define NEAR
#endif

#ifdef FP_OFF
#undef FP_OFF
#endif
#ifdef FP_SEG
#undef FP_SEG
#endif
#ifdef MK_FP
#undef MK_FP
#endif
#define FP_OFF(fp) ((unsigned int)(fp))
#define FP_SEG(fp) ((unsigned int)((unsigned long)(fp) >> 16))
#define MK_FP(sg,of) ((void far *)(((unsigned long)(sg) << 16)|(unsigned)(of)))

typedef struct {
	char           d_buf[21];
	char           d_attribute;
	unsigned short d_time;
	unsigned short d_date;
	long           d_size;
	char           d_name[13];
} DTA_BUF;

static DTA_BUF    dtabuf;
static union REGS reg, nreg;

#ifdef LDATA
static struct SREGS sreg;
#endif

static char ERNEM[] = "not enought mem\r\n";
#if GLOB
static char ERMISSCURLY[] = /*{*/ "Missing }.\r\n";
static char ERMISSBRACE[] = /*[*/ "Missing ].\r\n";
#endif
/***************************************************************/
static void NEAR error(char * s)
{
	write(2, "xargs: ", 7);
	write(2, s, strlen(s));
	exit(1);
}

static void NEAR set_dta(void)
{
	reg.h.ah = 0x1A;
#ifdef LDATA
	reg.x.dx = FP_OFF(&dtabuf);
	sreg.ds = FP_SEG(&dtabuf);
	intdosx(&reg, &nreg, &sreg);
#else
	reg.x.dx = (int) &dtabuf;
	intdos(&reg, &nreg);
#endif
}

static char * NEAR find_file(char *dir)
{
	if (dir != (char *) 0) { /* get first entry */
		set_dta();
		reg.h.ah = 0x4E;
		reg.h.cl = ArgsAttr;
#ifdef LDATA
		reg.x.dx = FP_OFF(dir);
		sreg.ds = FP_SEG(dir);
#else
		reg.x.dx = (unsigned) dir;
#endif
	} else {			/* get next entry */
#if HIDDEN_DOT && (ArgsAttr & 0x02) == 0
retry:;
#endif
		reg.h.ah = 0x4F;
#ifdef LDATA
		reg.x.dx = FP_OFF(&dtabuf);
		sreg.ds = FP_SEG(&dtabuf);
#else
		reg.x.dx = (unsigned) &dtabuf;
#endif
	}
#ifdef LDATA
	intdosx(&reg, &nreg, &sreg);
#else
	intdos(&reg, &nreg);
#endif

#ifdef HAVE_CFLAG
	if (nreg.x.cflag)
#else
	if (nreg.x.flags & 1)
#endif
		return (char *) 0;

	if (dtabuf.d_name[0] == 5) dtabuf.d_name[0] = 0xE5;
#if HIDDEN_DOT && (ArgsAttr & 0x02) == 0
	if (dtabuf.d_name[0] == '.') goto retry;
#endif
	return dtabuf.d_name;
}

#if CVLOWER
static char * NEAR jstrtolwr(char *p)
{
	char *op = p;

	while (*p) {
		if (isk1(*p) && p[1])
			p++;
		else if (('A' <= *p) && (*p <= 'Z'))
			*p |= 0x20;
		p++;
	}
	return op;
}
#else
#define jstrtolwr(s)	(s)
#endif

static char * NEAR jrindex(char * p, char c)
{
	char *oldp;

	for (oldp = (char *)0; *p; p++) {
		if (isk1(*p) && isk2(p[1])) {
			p++;
		} else if (*p == c) {
			oldp = p;
		}
	}
	return oldp;
}

#if DO_SORT
static int ustrcmp(s1, s2)
unsigned char **s1;
unsigned char **s2;
{
	unsigned char * p1 = *s1;
	unsigned char * p2 = *s2;

	while ((*p1 != '\0') && (*p1 == *p2)) {
		p1++;
		p2++;
	}
	return *p1 - *p2;
}
#endif

static char * NEAR xalloc(size_t n)
{
	char  *bp;

	if ((bp = calloc(1, n)) == (char *) 0)
		error(ERNEM);
	return bp;
}

static char * NEAR xrealloc(void *blk, size_t n)
{
	char  *bp;

	if ((bp = realloc(blk, n)) == (char *) 0)
		error(ERNEM);
	return bp;
}

#if GLOB
#if IGNORE_CASE
#define CaseFold(c)	if (Isupper(c)) {c = Tolower(c);}
#else
#define CaseFold(c)
#endif

#define Scan(c, p) {\
		c = UCH(*p++);\
		if (isk1(c) && isk2(*p)) c = (c << 8) + UCH(*p++);\
		CaseFold(c);\
	}

static int NEAR glob_match(char *t, char *p)
{
	int          r, include, exclude;
	unsigned int ks, ke, kp, kt;
	char         *s, *u;

	while (*p) {
		switch (*p) {
		case L_CURLY:
			if (is_end(*t)) return 0;
			u = s = t;
			p++;
			do {
				t = s;
				r = 1;
				do {
					if (is_end(*p)) error(ERMISSCURLY);
					Scan(kt, t);
					Scan(kp, p);
					if (kt != kp) r = 0;
				} while (*p != ',' && *p != R_CURLY);

				if (r == 1 && t - s > u - s) u = t;
				if (*p == ',') p++;
			} while (*p != R_CURLY);

			if (u == s) return 0;
			p++;
			t = u;
			break;

		case '?':
			if (is_end(*t)) return 0;
			if (isk1(*t) && isk2(*(t + 1))) t++;
			t++;
			p++;
			break;

		case '*':
			while ((r = glob_match(t, p + 1)) == 0 && !is_end(*t)) {
				if (isk1(*t) && isk2(*(t + 1))) t++;
				t++;
			}
			return r;

		case '[':
			if (is_end(*t)) return 0;
			include = exclude = 0;
			p++;
			if (*p == '^') {
				exclude = 1;
				p++;
			}
			Scan(kt, t);
			do {
				if (is_end(*p)) error(ERMISSBRACE);
				Scan(ks, p);
				ke = ks;
				if (*p == '-' && *(p + 1) != ']') {
					p++;
					if (is_end(*p)) error(ERMISSBRACE);
					Scan(ke, p);
				}
				if (ks <= kt && kt <= ke) {
					include = 1;
					/* no break to checking missing brace */
					/*break;*/
				}
			} while (*p != ']');

			if (include == exclude) return 0;
			p++;
			break;
#if DOT_DLM
		case '.':
			if (!is_end(*t)) return 0;
			if (*t == '.') t++;
			p++;
			break;
#endif
		default:
			Scan(kt, t);
			Scan(kp, p);
			if (kt != kp) return 0;
		}
	}
	return *t == '\0' ? 1 : 0;
}
#endif /* GLOB */

char  **
glob_filename(char *p)
{
	char *ap, *ap1, *ap2;
	char **res;
	char *path, *pat;
	char *wld;
	int   ac, plen;

	ap1 = jrindex(p, '\\');
	ap2 = jrindex(p, '/');
	if ((ap = (ap1 < ap2) ? ap2:ap1) == (char *)0)
		ap = jrindex(p, ':');

	if (ap) {
		pat = xalloc(strlen(ap + 1) + 1);
		strcpy(pat, ap + 1);
		ap2 = path = xalloc(strlen(p));
		for (ap1 = p; ap1 <= ap; ap1++) *ap2++ = *ap1;
		*ap2 = '\0';
	} else {
		pat = strdup(p);
		path = strdup("");
	}
#if GLOB
	wld = xalloc(strlen(path) + 4);
	strcat(strcpy(wld, path), "*.*");
#else
	wld = strdup(p);
#endif
	res = (char **)xalloc(sizeof(char *) * 2);
	*res = 0;
	ac = 0;
	plen = strlen(path);
	if ((ap = find_file(wld)) != (char *)0) {
		do {
#if GLOB
			if (glob_match(ap, pat))
#else
			/* always match */
#endif
			{
				strcat(strcpy(res[ac++] = xalloc(plen + strlen(ap) + 1), path), jstrtolwr(ap));
				res[ac] = 0;
				res = (char **) xrealloc(res, sizeof(char *) * (ac + 2));
			}
		} while ((ap = find_file((char *)0)) != (char *)0);
	}
	if (ac) {
#if DO_SORT
		qsort(res, ac, sizeof(res[0]), ustrcmp);
#endif
	} else {
		res[0] = strdup(p);
		res[1] = (char *)0;
	}
	free(pat);
	free(path);
	free(wld);
	return res;
}

static void NEAR setargv(char * NCptr, int *_ARGC, char ***_ARGV)
{
	int      i, jf;
	char     c;
	char     quote, withwild;	/* flags */
	char    *str;
	char    *ap;
	char    *cp;
	size_t   absiz;
	char    *buf = 0;
	size_t   bfsiz;
#define argbuf (*_ARGV)
#define argcnt (*_ARGC)
	char **res;

	buf = (char *) xalloc((bfsiz = REALLOC_UNIT * 2) * sizeof(buf[0]));
	argbuf = (char **) xalloc((absiz = REALLOC_UNIT) * sizeof(argbuf[0]));

	for (str = NCptr, i = 0;
	     ((c = *str) != '\r' && c != '\0');
	     str++, i++);
	*str = 0;
	cp = str;
	str = NCptr;
	for (; *str == ' ' || *str == '\t'; str++)
		;
	for (; --cp >= str && (*cp == ' ' || *cp == '\t' || *cp == '\r');)
		;
	cp[1] = 0;
	argcnt = 0;

	while (*str) {
		withwild = quote = 0;
		while (Isspace(*str)) str++;
		if (*str == 0) break;
		ap = buf;
		jf = 0;
		do {
			if (ap == &buf[bfsiz - 1]) {
				buf = (char *)xrealloc(buf, (bfsiz + REALLOC_UNIT) * sizeof(buf[0]));
				ap = &buf[bfsiz - 1];
				bfsiz += REALLOC_UNIT;
			}
			if (jf || (isk1(*str) && str[1])) {
				*ap++ = *str++;
				jf = !jf;
				continue;
			}
			if (!quote && iswildchar(*str))
				withwild = 1;
			if (*str == quote) {	/* end of quoted string */
				quote = 0;
				str++;	/* Skip quote */
			} else if (!quote && (*str == '\'' || *str == '"')) {
				quote = *str++;
			} else {
				 if (*str == '\\' && (str[1] == '"' || str[1] == '\''))
					str++;
				 *ap++ = *str++;
			}
		} while (*str && (quote || !Isspace(*str)));
		*ap = 0;
		res = (char **)0;
		if (withwild) {
			res = glob_filename(buf);
		}
		if (res == (char **)0) {
			res = (char **)xalloc(sizeof(char *) * 2);
			res[0] = strdup(buf);
			res[1] = (char *)0;
		}
		for (i = 0; res[i]; i++) {
			argbuf[argcnt++] = res[i];
			if ((size_t) argcnt >= absiz) {
				argbuf = (char **) xrealloc(argbuf, (absiz += REALLOC_UNIT) * sizeof(argbuf[0]));
			}
		}
		free(res);
	}
	argbuf[argcnt] = (char *) 0;
	argbuf = (char **) xrealloc(argbuf, (argcnt + 1) * sizeof(argbuf[0]));
	free(buf);
}

void
xargs(int *argc, char ***argv)
{
	char far *pp;
	char *s;
	int  i;

	if ((s = getenv("CMDLINE")) == (char *)0) {
		s = xalloc(256);
		pp = MK_FP(PSP, 129);
		jstrtolwr(argv[0][0]);
		for (i = 0; i < 127 && (s[i] = argv[0][0][i]) != '\0'; i++)
			;
		if (*pp != ' ' && *pp != '\t') s[i++] = ' '; /* better way ? */
		for ( ; i < 127 && (s[i] = *pp++) != '\0' && s[i] != '\r'; i++)
			;
		if (i == 127 && (s[126] != '\r' && s[126] != 0)) {
			s[126] = 0;
		}
	}
	setargv(s, argc, argv);
}

#ifdef TEST
#include <stdio.h>
#include <stdlib.h>

static char QUOTE[]  = "\033[33m\"\033[m";

void
main(argc, argv)
int    argc;
char **argv;
{
	int a = 0;

	(void)setbuf(stdout, malloc(BUFSIZ));	/* for speed up */

	xargs(&argc, &argv);

	printf("[argv]-------------------------- %d argument(s)\n", argc);
	while (argc--) {
		printf("[%4d]%s%s%s\n", a++, QUOTE, *argv++, QUOTE);
	}
	printf("-------------------------------- Thats all\n");
}
#endif
